"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.bridge as bridge
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class BridgeTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.bridge
    """

    def setup_loader_modules(self):
        return {bridge: {}}

    def test_show(self):
        """
        Test for Returns bridges interfaces
        along with enslaved physical interfaces
        """
        mock = MagicMock(return_value=True)
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertTrue(bridge.show("br"))

    def test_list_(self):
        """
        Test for Returns the machine's bridges list
        """
        mock = MagicMock(return_value=None)
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.list_(), None)

        mock = MagicMock(return_value=["A", "B"])
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.list_(), ["A", "B"])

    def test_interfaces(self):
        """
        Test for Returns interfaces attached to a bridge
        """
        self.assertEqual(bridge.interfaces(), None)

        mock = MagicMock(return_value={"interfaces": "A"})
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.interfaces("br"), "A")

    def test_find_interfaces(self):
        """
        Test for Returns the bridge to which the interfaces are bond to
        """
        mock = MagicMock(return_value=None)
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.find_interfaces(), None)

        mock = MagicMock(return_value={"interfaces": "A"})
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.find_interfaces(), {})

    def test_add(self):
        """
        Test for Creates a bridge
        """
        mock = MagicMock(return_value="A")
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.add(), "A")

    def test_delete(self):
        """
        Test for Deletes a bridge
        """
        mock = MagicMock(return_value="A")
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.delete(), "A")

    def test_addif(self):
        """
        Test for Adds an interface to a bridge
        """
        mock = MagicMock(return_value="A")
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.addif(), "A")

    def test_delif(self):
        """
        Test for Removes an interface from a bridge
        """
        mock = MagicMock(return_value="A")
        with patch.object(bridge, "_os_dispatch", mock):
            self.assertEqual(bridge.delif(), "A")

    def test_stp(self):
        """
        Test for Sets Spanning Tree Protocol state for a bridge
        """
        with patch.dict(bridge.__grains__, {"kernel": "Linux"}):
            mock = MagicMock(return_value="Linux")
            with patch.object(bridge, "_os_dispatch", mock):
                self.assertEqual(bridge.stp(), "Linux")

        with patch.dict(bridge.__grains__, {"kernel": "FreeBSD"}):
            mock = MagicMock(return_value="FreeBSD")
            with patch.object(bridge, "_os_dispatch", mock):
                self.assertEqual(bridge.stp(), "FreeBSD")

        with patch.dict(bridge.__grains__, {"kernel": None}):
            self.assertFalse(bridge.stp())



import ctypes
import logging
import multiprocessing
import time
from concurrent.futures.thread import ThreadPoolExecutor

import pytest
import salt.config
import salt.exceptions
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.log.setup
import salt.master
import salt.transport.client
import salt.transport.server
import salt.transport.zeromq
import salt.utils.platform
import salt.utils.process
import salt.utils.stringutils
import zmq.eventloop.ioloop
from saltfactories.utils.processes import terminate_process
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


class Collector(salt.utils.process.SignalHandlingProcess):
    def __init__(
        self, minion_config, pub_uri, aes_key, timeout=30, zmq_filtering=False
    ):
        super().__init__()
        self.minion_config = minion_config
        self.pub_uri = pub_uri
        self.aes_key = aes_key
        self.timeout = timeout
        self.hard_timeout = time.time() + timeout + 30
        self.manager = multiprocessing.Manager()
        self.results = self.manager.list()
        self.zmq_filtering = zmq_filtering
        self.stopped = multiprocessing.Event()
        self.started = multiprocessing.Event()
        self.running = multiprocessing.Event()

    def _rotate_secrets(self, now=None):
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "serial": multiprocessing.Value(
                ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
            ),
            "reload": salt.crypt.Crypticle.generate_key_string,
            "rotate_master_key": self._rotate_secrets,
        }

    def run(self):
        """
        Gather results until then number of seconds specified by timeout passes
        without receiving a message
        """
        ctx = zmq.Context()
        sock = ctx.socket(zmq.SUB)
        sock.setsockopt(zmq.LINGER, -1)
        sock.setsockopt(zmq.SUBSCRIBE, b"")
        sock.connect(self.pub_uri)
        last_msg = time.time()
        crypticle = salt.crypt.Crypticle(self.minion_config, self.aes_key)
        self.started.set()
        while True:
            curr_time = time.time()
            if time.time() > self.hard_timeout:
                break
            if curr_time - last_msg >= self.timeout:
                break
            try:
                payload = sock.recv(zmq.NOBLOCK)
            except zmq.ZMQError:
                time.sleep(0.01)
            else:
                try:
                    serial_payload = salt.payload.loads(payload)
                    payload = crypticle.loads(serial_payload["load"])
                    if not payload:
                        continue
                    if "start" in payload:
                        self.running.set()
                        continue
                    if "stop" in payload:
                        break
                    last_msg = time.time()
                    self.results.append(payload["jid"])
                except salt.exceptions.SaltDeserializationError:
                    if not self.zmq_filtering:
                        log.exception("Failed to deserialize...")
                        break

    def __enter__(self):
        self.manager.__enter__()
        self.start()
        # Wait until we can start receiving events
        self.started.wait()
        self.started.clear()
        return self

    def __exit__(self, *args):
        # Wait until we either processed all expected messages or we reach the hard timeout
        join_secs = self.hard_timeout - time.time()
        log.info("Waiting at most %s seconds before exiting the collector", join_secs)
        self.join(join_secs)
        self.terminate()
        # Cast our manager.list into a plain list
        self.results = list(self.results)
        # Terminate our multiprocessing manager
        self.manager.__exit__(*args)
        log.debug("The collector has exited")
        self.stopped.set()


class PubServerChannelProcess(salt.utils.process.SignalHandlingProcess):
    def __init__(self, master_config, minion_config, **collector_kwargs):
        super().__init__()
        self._closing = False
        self.master_config = master_config
        self.minion_config = minion_config
        self.collector_kwargs = collector_kwargs
        self.aes_key = salt.crypt.Crypticle.generate_key_string()
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(self.aes_key),
            ),
            "serial": multiprocessing.Value(
                ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
            ),
        }
        self.process_manager = salt.utils.process.ProcessManager(
            name="ZMQ-PubServer-ProcessManager"
        )
        self.pub_server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(
            self.master_config
        )
        self.pub_server_channel.pre_fork(
            self.process_manager,
            kwargs={"log_queue": salt.log.setup.get_multiprocessing_logging_queue()},
        )
        self.pub_uri = "tcp://{interface}:{publish_port}".format(**self.master_config)
        self.queue = multiprocessing.Queue()
        self.stopped = multiprocessing.Event()
        self.collector = Collector(
            self.minion_config, self.pub_uri, self.aes_key, **self.collector_kwargs
        )

    def run(self):
        try:
            while True:
                payload = self.queue.get()
                if payload is None:
                    log.debug("We received the stop sentinel")
                    break
                self.pub_server_channel.publish(payload)
        except KeyboardInterrupt:
            pass
        finally:
            self.stopped.set()

    def _handle_signals(self, signum, sigframe):
        self.close()
        super()._handle_signals(signum, sigframe)

    def close(self):
        if self._closing:
            return
        self._closing = True
        if self.process_manager is None:
            return
        self.process_manager.terminate()
        self.pub_server_channel.pub_close()
        # Really terminate any process still left behind
        for pid in self.process_manager._process_map:
            terminate_process(pid=pid, kill_children=True, slow_stop=False)
        self.process_manager = None

    def publish(self, payload):
        self.queue.put(payload)

    def __enter__(self):
        self.start()
        self.collector.__enter__()
        attempts = 30
        while attempts > 0:
            self.publish({"tgt_type": "glob", "tgt": "*", "jid": -1, "start": True})
            if self.collector.running.wait(1) is True:
                break
            attempts -= 1
        else:
            pytest.fail("Failed to confirm the collector has started")
        return self

    def __exit__(self, *args):
        # Publish a payload to tell the collection it's done processing
        self.publish({"tgt_type": "glob", "tgt": "*", "jid": -1, "stop": True})
        # Now trigger the collector to also exit
        self.collector.__exit__(*args)
        # We can safely wait here without a timeout because the Collector instance has a
        # hard timeout set, so eventually Collector.stopped will be set
        self.collector.stopped.wait()
        # Stop our own processing
        self.queue.put(None)
        # Wait at most 10 secs for the above `None` in the queue to be processed
        self.stopped.wait(10)
        self.close()
        self.terminate()
        log.info("The PubServerChannelProcess has terminated")


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_publish_to_pubserv_ipc(salt_master, salt_minion):
    """
    Test sending 10K messags to ZeroMQPubServerChannel using IPC transport

    ZMQ's ipc transport not supported on Windows
    """
    opts = dict(salt_master.config.copy(), ipc_mode="ipc", pub_hwm=0)
    with PubServerChannelProcess(opts, salt_minion.config.copy()) as server_channel:
        send_num = 10000
        expect = []
        for idx in range(send_num):
            expect.append(idx)
            load = {"tgt_type": "glob", "tgt": "*", "jid": idx}
            server_channel.publish(load)
    results = server_channel.collector.results
    assert len(results) == send_num, "{} != {}, difference: {}".format(
        len(results), send_num, set(expect).difference(results)
    )


@pytest.mark.skip_on_freebsd
@pytest.mark.slow_test
def test_issue_36469_tcp(salt_master, salt_minion):
    """
    Test sending both large and small messags to publisher using TCP

    https://github.com/saltstack/salt/issues/36469
    """

    def _send_small(opts, sid, num=10):
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        for idx in range(num):
            load = {"tgt_type": "glob", "tgt": "*", "jid": "{}-s{}".format(sid, idx)}
            server_channel.publish(load)
        time.sleep(0.3)
        server_channel.close_pub()

    def _send_large(opts, sid, num=10, size=250000 * 3):
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        for idx in range(num):
            load = {
                "tgt_type": "glob",
                "tgt": "*",
                "jid": "{}-l{}".format(sid, idx),
                "xdata": "0" * size,
            }
            server_channel.publish(load)
        time.sleep(0.3)
        server_channel.close_pub()

    opts = dict(salt_master.config.copy(), ipc_mode="tcp", pub_hwm=0)
    send_num = 10 * 4
    expect = []
    with PubServerChannelProcess(opts, salt_minion.config.copy()) as server_channel:
        assert "aes" in salt.master.SMaster.secrets
        with ThreadPoolExecutor(max_workers=4) as executor:
            executor.submit(_send_small, opts, 1)
            executor.submit(_send_large, opts, 2)
            executor.submit(_send_small, opts, 3)
            executor.submit(_send_large, opts, 4)
        expect.extend(["{}-s{}".format(a, b) for a in range(10) for b in (1, 3)])
        expect.extend(["{}-l{}".format(a, b) for a in range(10) for b in (2, 4)])
    results = server_channel.collector.results
    assert len(results) == send_num, "{} != {}, difference: {}".format(
        len(results), send_num, set(expect).difference(results)
    )


@pytest.mark.skip_on_windows
@pytest.mark.slow_test
def test_zeromq_filtering(salt_master, salt_minion):
    """
    Test sending messages to publisher using UDP with zeromq_filtering enabled
    """
    opts = dict(
        salt_master.config.copy(),
        ipc_mode="ipc",
        pub_hwm=0,
        zmq_filtering=True,
        acceptance_wait_time=5,
    )
    send_num = 1
    expect = []
    with patch(
        "salt.utils.minions.CkMinions.check_minions",
        MagicMock(
            return_value={
                "minions": [salt_minion.id],
                "missing": [],
                "ssh_minions": False,
            }
        ),
    ):
        with PubServerChannelProcess(
            opts, salt_minion.config.copy(), zmq_filtering=True
        ) as server_channel:
            expect.append(send_num)
            load = {"tgt_type": "glob", "tgt": "*", "jid": send_num}
            server_channel.publish(load)
        results = server_channel.collector.results
        assert len(results) == send_num, "{} != {}, difference: {}".format(
            len(results), send_num, set(expect).difference(results)
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Copyright 2016 VMware, Inc.  All rights reserved.

This module defines basic helper functions used in the sampe codes
"""

# pylint: skip-file
__author__ = 'VMware, Inc'

from pyVmomi import vim, vmodl, SoapStubAdapter
#import the VSAN API python bindings
import vsanmgmtObjects

VSAN_API_VC_SERVICE_ENDPOINT = '/vsanHealth'
VSAN_API_ESXI_SERVICE_ENDPOINT = '/vsan'

#Constuct a stub for VSAN API access using VC or ESXi sessions from  existing
#stubs. Correspoding VC or ESXi service endpoint is required. VC service
#endpoint is used as default
def _GetVsanStub(
      stub, endpoint=VSAN_API_VC_SERVICE_ENDPOINT,
      context=None, version='vim.version.version10'
   ):

   hostname = stub.host.split(':')[0]
   vsanStub = SoapStubAdapter(
      host=hostname,
      path=endpoint,
      version=version,
      sslContext=context
   )
   vsanStub.cookie = stub.cookie
   return vsanStub

#Construct a stub for access VC side VSAN APIs
def GetVsanVcStub(stub, context=None):
   return _GetVsanStub(stub, endpoint=VSAN_API_VC_SERVICE_ENDPOINT,
                       context=context)

#Construct a stub for access ESXi side VSAN APIs
def GetVsanEsxStub(stub, context=None):
   return _GetVsanStub(stub, endpoint=VSAN_API_ESXI_SERVICE_ENDPOINT,
                       context=context)

#Construct a stub for access ESXi side VSAN APIs
def GetVsanVcMos(vcStub, context=None):
   vsanStub = GetVsanVcStub(vcStub, context)
   vcMos = {
      'vsan-disk-management-system' : vim.cluster.VsanVcDiskManagementSystem(
                                         'vsan-disk-management-system',
                                         vsanStub
                                      ),
      'vsan-stretched-cluster-system' : vim.cluster.VsanVcStretchedClusterSystem(
                                           'vsan-stretched-cluster-system',
                                           vsanStub
                                        ),
      'vsan-cluster-config-system' : vim.cluster.VsanVcClusterConfigSystem(
                                        'vsan-cluster-config-system',
                                        vsanStub
                                     ),
      'vsan-performance-manager' : vim.cluster.VsanPerformanceManager(
                                      'vsan-performance-manager',
                                      vsanStub
                                   ),
      'vsan-cluster-health-system' : vim.cluster.VsanVcClusterHealthSystem(
                                        'vsan-cluster-health-system',
                                        vsanStub
                                     ),
      'vsan-upgrade-systemex' : vim.VsanUpgradeSystemEx(
                                   'vsan-upgrade-systemex',
                                    vsanStub
                                ),
      'vsan-cluster-space-report-system' : vim.cluster.VsanSpaceReportSystem(
                                              'vsan-cluster-space-report-system',
                                              vsanStub
                                           ),

      'vsan-cluster-object-system' : vim.cluster.VsanObjectSystem(
                                        'vsan-cluster-object-system',
                                        vsanStub
                                     ),
   }

   return vcMos

#Construct a stub for access ESXi side VSAN APIs
def GetVsanEsxMos(esxStub, context=None):
   vsanStub = GetVsanEsxStub(esxStub, context)
   esxMos = {
      'vsan-performance-manager' : vim.cluster.VsanPerformanceManager(
                                      'vsan-performance-manager',
                                      vsanStub
                                   ),
      'ha-vsan-health-system' : vim.host.VsanHealthSystem(
                                        'ha-vsan-health-system',
                                        vsanStub
                                     ),
      'vsan-object-system' : vim.cluster.VsanObjectSystem(
                                        'vsan-object-system',
                                        vsanStub
                                     ),
   }

   return esxMos

#Convert a VSAN Task to a Task MO binding to VC service
#@param vsanTask the VSAN Task MO
#@param stub the stub for the VC API
def ConvertVsanTaskToVcTask(vsanTask, vcStub):
  vcTask = vim.Task(vsanTask._moId, vcStub)
  return vcTask

def WaitForTasks(tasks, si):
   """
   Given the service instance si and tasks, it returns after all the
   tasks are complete
   """

   pc = si.content.propertyCollector

   taskList = [str(task) for task in tasks]

   # Create filter
   objSpecs = [vmodl.query.PropertyCollector.ObjectSpec(obj=task)
                                                            for task in tasks]
   propSpec = vmodl.query.PropertyCollector.PropertySpec(type=vim.Task,
                                                         pathSet=[], all=True)
   filterSpec = vmodl.query.PropertyCollector.FilterSpec()
   filterSpec.objectSet = objSpecs
   filterSpec.propSet = [propSpec]
   filter = pc.CreateFilter(filterSpec, True)

   try:
      version, state = None, None

      # Loop looking for updates till the state moves to a completed state.
      while len(taskList):
         update = pc.WaitForUpdates(version)
         for filterSet in update.filterSet:
            for objSet in filterSet.objectSet:
               task = objSet.obj
               for change in objSet.changeSet:
                  if change.name == 'info':
                     state = change.val.state
                  elif change.name == 'info.state':
                     state = change.val
                  else:
                     continue

                  if not str(task) in taskList:
                     continue

                  if state == vim.TaskInfo.State.success:
                     # Remove task from taskList
                     taskList.remove(str(task))
                  elif state == vim.TaskInfo.State.error:
                     raise task.info.error
         # Move to next version
         version = update.version
   finally:
      if filter:
         filter.Destroy()



# pylint: disable=E8231
# Salt libs
import pytest
import salt.beacons.salt_monitor as salt_monitor

TEST_CONFIG = [
    {
        "config": [{"salt_fun": ["test.ping", "test.version"]}],
        "expected_validate": (True, "valid config"),
        "expected_beacon": [
            {"salt_fun": "test.ping", "ret": True},
            {"salt_fun": "test.version", "ret": "3000"},
        ],
    },
    {
        "config": [
            {
                "salt_fun": [
                    {
                        "cmd.run": {
                            "args": [
                                "echo hello world",
                            ]
                        }
                    },
                    "test.version",
                ]
            }
        ],  # *args behaves weird on list with single string, does it happen in yaml?
        "expected_validate": (True, "valid config"),
        "expected_beacon": [
            {
                "salt_fun": "cmd.run",
                "ret": (("echo hello world",), {}),
                "args": ["echo hello world"],
            },
            {"salt_fun": "test.version", "ret": "3000"},
        ],
    },
    {
        "config": [
            {
                "salt_fun": [
                    {
                        "cmd.run": {
                            "args": [
                                "echo hello world",
                            ],
                            "kwargs": [{"shell": "ps"}],
                        }
                    }
                ]
            }
        ],
        "expected_validate": (True, "valid config"),
        "expected_beacon": [
            {
                "salt_fun": "cmd.run",
                "ret": (("echo hello world",), {"shell": "ps"}),
                "args": ["echo hello world"],
                "kwargs": {"shell": "ps"},
            }
        ],
    },
    {
        "config": [
            {"salt_fun": [{"cmd.run": {"args": "echo hello world"}}, "test.version"]}
        ],
        "expected_validate": (False, "args key for fun cmd.run must be list"),
        "expected_beacon": None,  # None != []
    },
    {
        "config": [
            {
                "salt_fun": [
                    {
                        "cmd.run": {
                            "args": ["echo hello world"],
                            "kwargs": {"shell": "ps"},
                        }
                    }
                ]
            }
        ],
        "expected_validate": (
            False,
            "kwargs key for fun cmd.run must be list of key value pairs",
        ),
        "expected_beacon": None,
    },
    {
        "config": [
            {
                "salt_fun": [
                    {"cmd.run": {"args": ["echo hello world"], "kwargs": ["shell=ps"]}}
                ]
            }
        ],
        "expected_validate": (
            False,
            "{} is not a key / value pair".format("shell=ps"),
        ),
        "expected_beacon": None,
    },
    {
        "config": [
            {
                "salt_fun": [
                    {
                        "cmd.run": {
                            "args": ["echo hello world"],
                            "kwargs": [{"shell": "ps"}],
                            "bg": True,
                        }
                    }
                ]
            }
        ],
        "expected_validate": (False, "key bg not allowed under fun cmd.run"),
        "expected_beacon": None,
    },
    {
        "config": [
            {
                "salt_fun": [
                    {"bogus.fun": {"args": ["echo hello world"]}},
                    "test.version",
                ]
            }
        ],
        "expected_validate": (False, "bogus.fun not in __salt__"),
        "expected_beacon": None,
    },
    {
        "config": [{"salt_fun": ["test.ping", "test.false"]}],
        "expected_validate": (True, "valid config"),
        "expected_beacon": [{"salt_fun": "test.ping", "ret": True}],
    },
    {
        "config": [{"salt_fun": ["test.false"]}],
        "expected_validate": (True, "valid config"),
        "expected_beacon": [],
    },
    {
        "config": [{"salt_fun": "test.ping"}],
        "expected_validate": (True, "valid config"),
        "expected_beacon": [{"salt_fun": "test.ping", "ret": True}],
    },
]


def mock_test_ping():
    return True


def mock_test_version():
    return "3000"


def mock_test(*args, **kwargs):
    return args, kwargs


def mock_test_false():
    return False


@pytest.fixture
def configure_loader_modules():
    return {
        salt_monitor: {
            "__salt__": {
                "test.ping": mock_test_ping,
                "test.version": mock_test_version,
                "cmd.run": mock_test,
                "test.false": mock_test_false,
            }
        }
    }


def test_validate():
    for i, config in enumerate(TEST_CONFIG):
        valid = salt_monitor.validate(config["config"])
        assert valid == config["expected_validate"]


def test_beacon():
    for config in TEST_CONFIG:
        if config["expected_beacon"] is None:
            continue
        events = salt_monitor.beacon(config["config"])
        assert events == config["expected_beacon"]



"""
Tests for the fileserver runner
"""

import contextlib
import pathlib

import pytest
from tests.support.case import ShellCase
from tests.support.helpers import PRE_PYTEST_SKIP_REASON
from tests.support.mock import MagicMock, create_autospec, patch
from tests.support.runtests import RUNTIME_VARS


@pytest.mark.windows_whitelisted
class FileserverTest(ShellCase):
    """
    Test the fileserver runner
    """

    @pytest.mark.slow_test
    def test_dir_list(self):
        """
        fileserver.dir_list
        """
        ret = self.run_run_plus(fun="fileserver.dir_list")
        self.assertIsInstance(ret["return"], list)
        self.assertTrue("_modules" in ret["return"])

        # Backend submitted as a string
        ret = self.run_run_plus(fun="fileserver.dir_list", backend="roots")
        self.assertIsInstance(ret["return"], list)
        self.assertTrue("_modules" in ret["return"])

        # Backend submitted as a list
        ret = self.run_run_plus(fun="fileserver.dir_list", backend=["roots"])
        self.assertIsInstance(ret["return"], list)
        self.assertTrue("_modules" in ret["return"])

    @pytest.mark.slow_test
    def test_empty_dir_list(self):
        """
        fileserver.empty_dir_list
        """
        ret = self.run_run_plus(fun="fileserver.empty_dir_list")
        self.assertIsInstance(ret["return"], list)
        self.assertEqual(ret["return"], [])

        # Backend submitted as a string
        ret = self.run_run_plus(fun="fileserver.empty_dir_list", backend="roots")
        self.assertIsInstance(ret["return"], list)
        self.assertEqual(ret["return"], [])

        # Backend submitted as a list
        ret = self.run_run_plus(fun="fileserver.empty_dir_list", backend=["roots"])
        self.assertIsInstance(ret["return"], list)
        self.assertEqual(ret["return"], [])

    @pytest.mark.slow_test
    def test_envs(self):
        """
        fileserver.envs
        """
        ret = self.run_run_plus(fun="fileserver.envs")
        self.assertIsInstance(ret["return"], list)

        # Backend submitted as a string
        ret = self.run_run_plus(fun="fileserver.envs", backend="roots")
        self.assertIsInstance(ret["return"], list)

        # Backend submitted as a list
        ret = self.run_run_plus(fun="fileserver.envs", backend=["roots"])
        self.assertIsInstance(ret["return"], list)

    @pytest.mark.slow_test
    def test_clear_file_list_cache(self):
        """
        fileserver.clear_file_list_cache

        If this test fails, then something may have changed in the test suite
        and we may have more than just "roots" configured in the fileserver
        backends. This assert will need to be updated accordingly.
        """
        saltenvs = sorted(self.run_run_plus(fun="fileserver.envs")["return"])

        @contextlib.contextmanager
        def gen_cache():
            """
            Create file_list cache so we have something to clear
            """
            for saltenv in saltenvs:
                self.run_run_plus(fun="fileserver.file_list", saltenv=saltenv)
            yield

        # Test with no arguments
        with gen_cache():
            ret = self.run_run_plus(fun="fileserver.clear_file_list_cache")
            ret["return"]["roots"].sort()
            self.assertEqual(ret["return"], {"roots": saltenvs})

        # Test with backend passed as string
        with gen_cache():
            ret = self.run_run_plus(
                fun="fileserver.clear_file_list_cache", backend="roots"
            )
            ret["return"]["roots"].sort()
            self.assertEqual(ret["return"], {"roots": saltenvs})

        # Test with backend passed as list
        with gen_cache():
            ret = self.run_run_plus(
                fun="fileserver.clear_file_list_cache", backend=["roots"]
            )
            ret["return"]["roots"].sort()
            self.assertEqual(ret["return"], {"roots": saltenvs})

        # Test with backend passed as string, but with a nonsense backend
        with gen_cache():
            ret = self.run_run_plus(
                fun="fileserver.clear_file_list_cache", backend="notarealbackend"
            )
            self.assertEqual(ret["return"], {})

        # Test with saltenv passed as string
        with gen_cache():
            ret = self.run_run_plus(
                fun="fileserver.clear_file_list_cache", saltenv="base"
            )
            self.assertEqual(ret["return"], {"roots": ["base"]})

        # Test with saltenv passed as list
        with gen_cache():
            ret = self.run_run_plus(
                fun="fileserver.clear_file_list_cache", saltenv=["base"]
            )
            self.assertEqual(ret["return"], {"roots": ["base"]})

        # Test with saltenv passed as string, but with a nonsense saltenv
        with gen_cache():
            ret = self.run_run_plus(
                fun="fileserver.clear_file_list_cache", saltenv="notarealsaltenv"
            )
            self.assertEqual(ret["return"], {})

        # Test with both backend and saltenv passed
        with gen_cache():
            ret = self.run_run_plus(
                fun="fileserver.clear_file_list_cache", backend="roots", saltenv="base"
            )
            self.assertEqual(ret["return"], {"roots": ["base"]})

    @pytest.mark.slow_test
    def test_file_list(self):
        """
        fileserver.file_list
        """
        ret = self.run_run_plus(fun="fileserver.file_list")
        self.assertIsInstance(ret["return"], list)
        self.assertTrue("grail/scene33" in ret["return"])

        # Backend submitted as a string
        ret = self.run_run_plus(fun="fileserver.file_list", backend="roots")
        self.assertIsInstance(ret["return"], list)
        self.assertTrue("grail/scene33" in ret["return"])

        # Backend submitted as a list
        ret = self.run_run_plus(fun="fileserver.file_list", backend=["roots"])
        self.assertIsInstance(ret["return"], list)
        self.assertTrue("grail/scene33" in ret["return"])

    @pytest.mark.slow_test
    @pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
    def test_symlink_list(self):
        """
        fileserver.symlink_list
        """
        source_sym = pathlib.Path(RUNTIME_VARS.TMP_BASEENV_STATE_TREE) / "source_sym_1"
        source_sym.write_text("")
        dest_sym = pathlib.Path(RUNTIME_VARS.TMP_BASEENV_STATE_TREE) / "dest_sym_1"
        dest_sym.symlink_to(str(source_sym))
        self.addCleanup(dest_sym.unlink)
        self.addCleanup(source_sym.unlink)

        ret = self.run_run_plus(fun="fileserver.symlink_list")
        self.assertIsInstance(ret["return"], dict)
        self.assertTrue("dest_sym_1" in ret["return"])

        # Backend submitted as a string
        ret = self.run_run_plus(fun="fileserver.symlink_list", backend="roots")
        self.assertIsInstance(ret["return"], dict)
        self.assertTrue("dest_sym_1" in ret["return"])

        # Backend submitted as a list
        ret = self.run_run_plus(fun="fileserver.symlink_list", backend=["roots"])
        self.assertIsInstance(ret["return"], dict)
        self.assertTrue("dest_sym_1" in ret["return"])

    @pytest.mark.slow_test
    def test_update(self):
        """
        fileserver.update
        """
        ret = self.run_run_plus(fun="fileserver.update")
        self.assertTrue(ret["return"])

        # Backend submitted as a string
        ret = self.run_run_plus(fun="fileserver.update", backend="roots")
        self.assertTrue(ret["return"])

        # Backend submitted as a list
        ret = self.run_run_plus(fun="fileserver.update", backend=["roots"])
        self.assertTrue(ret["return"])

        # Other arguments are passed to backend
        def mock_gitfs_update(remotes=None):
            pass

        mock_backend_func = create_autospec(mock_gitfs_update)
        mock_return_value = {
            "gitfs.envs": None,  # This is needed to activate the backend
            "gitfs.update": mock_backend_func,
        }
        with patch("salt.loader.fileserver", MagicMock(return_value=mock_return_value)):
            ret = self.run_run_plus(
                fun="fileserver.update", backend="gitfs", remotes="myrepo,yourrepo"
            )
            self.assertTrue(ret["return"])
            mock_backend_func.assert_called_once_with(remotes="myrepo,yourrepo")

        # Unknown arguments are passed to backend
        with patch("salt.loader.fileserver", MagicMock(return_value=mock_return_value)):
            ret = self.run_run_plus(
                fun="fileserver.update", backend="gitfs", unknown_arg="foo"
            )
        self.assertIn(
            "Passed invalid arguments: got an unexpected keyword argument"
            " 'unknown_arg'",
            ret["return"],
        )



"""
unit tests for the script engine
"""

import salt.config
import salt.engines.script as script
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class EngineScriptTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.engine.script
    """

    def setup_loader_modules(self):

        opts = salt.config.DEFAULT_MASTER_OPTS
        return {script: {"__opts__": opts}}

    def test__get_serializer(self):
        """
        Test known serializer is returned or exception is raised
        if unknown serializer
        """
        for serializers in ("json", "yaml", "msgpack"):
            self.assertTrue(script._get_serializer(serializers))

        with self.assertRaises(CommandExecutionError):
            script._get_serializer("bad")

    def test__read_stdout(self):
        """
        Test we can yield stdout
        """
        with patch("subprocess.Popen") as popen_mock:
            popen_mock.stdout.readline.return_value = "test"
            self.assertEqual(next(script._read_stdout(popen_mock)), "test")



import logging

import pytest
import salt.config
import salt.loader
import salt.modules.azurearm_dns as azurearm_dns
from tests.support.mock import MagicMock
from tests.support.sminion import create_sminion

HAS_LIBS = False
try:
    import azure.mgmt.dns.models  # pylint: disable=import-error

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(
        HAS_LIBS is False, reason="The azure.mgmt.dns module must be installed."
    ),
]


class AzureObjMock:
    """
    mock azure object for as_dict calls
    """

    args = None
    kwargs = None

    def __init__(self, args, kwargs, return_value=None):
        self.args = args
        self.kwargs = kwargs
        self.__return_value = return_value

    def __getattr__(self, item):
        return self

    def __call__(self, *args, **kwargs):
        return MagicMock(return_value=self.__return_value)()

    def as_dict(self, *args, **kwargs):
        return self.args, self.kwargs


class AzureFuncMock:
    """
    mock azure client function calls
    """

    def __init__(self, return_value=None):
        self.__return_value = return_value

    def __getattr__(self, item):
        return self

    def __call__(self, *args, **kwargs):
        return MagicMock(return_value=self.__return_value)()

    def create_or_update(self, *args, **kwargs):
        azure_obj = AzureObjMock(args, kwargs)
        return azure_obj


class AzureSubMock:
    """
    mock azure client sub-modules
    """

    record_sets = AzureFuncMock()
    zones = AzureFuncMock()

    def __init__(self, return_value=None):
        self.__return_value = return_value

    def __getattr__(self, item):
        return self

    def __call__(self, *args, **kwargs):
        return MagicMock(return_value=self.__return_value)()


class AzureClientMock:
    """
    mock azure client
    """

    def __init__(self, return_value=AzureSubMock):
        self.__return_value = return_value

    def __getattr__(self, item):
        return self

    def __call__(self, *args, **kwargs):
        return MagicMock(return_value=self.__return_value)()


@pytest.fixture
def credentials():
    azurearm_dns.__virtual__()
    return {
        "client_id": "CLIENT_ID",
        "secret": "SECRET",
        "subscription_id": "SUBSCRIPTION_ID",
        "tenant": "TENANT",
    }


@pytest.fixture
def configure_loader_modules():
    """
    setup loader modules and override the azurearm.get_client utility
    """
    minion_config = create_sminion().opts.copy()
    utils = salt.loader.utils(minion_config)
    funcs = salt.loader.minion_mods(
        minion_config, utils=utils, whitelist=["azurearm_dns", "config"]
    )
    utils["azurearm.get_client"] = AzureClientMock()
    return {
        azurearm_dns: {"__utils__": utils, "__salt__": funcs},
    }


def test_record_set_create_or_update(credentials):
    """
    tests record set object creation
    """
    expected = {
        "if_match": None,
        "if_none_match": None,
        "parameters": {"arecords": [{"ipv4_address": "10.0.0.1"}], "ttl": 300},
        "record_type": "A",
        "relative_record_set_name": "myhost",
        "resource_group_name": "testgroup",
        "zone_name": "myzone",
    }

    record_set_args, record_set_kwargs = azurearm_dns.record_set_create_or_update(
        "myhost",
        "myzone",
        "testgroup",
        "A",
        arecords=[{"ipv4_address": "10.0.0.1"}],
        ttl=300,
        **credentials
    )

    for key, val in record_set_kwargs.items():
        if isinstance(val, azure.mgmt.dns.models.RecordSet):
            record_set_kwargs[key] = val.as_dict()

    assert record_set_kwargs == expected


def test_zone_create_or_update(credentials):
    """
    tests zone object creation
    """
    expected = {
        "if_match": None,
        "if_none_match": None,
        "parameters": {"location": "global", "zone_type": "Public"},
        "resource_group_name": "testgroup",
        "zone_name": "myzone",
    }

    zone_args, zone_kwargs = azurearm_dns.zone_create_or_update(
        "myzone", "testgroup", **credentials
    )

    for key, val in zone_kwargs.items():
        if isinstance(val, azure.mgmt.dns.models.Zone):
            zone_kwargs[key] = val.as_dict()

    assert zone_kwargs == expected
